# $Header: /home/cvs/Torped/EasyPublisher16/lib/python/Products/EasyImage/EasyImage.py,v 1.18 2002/05/08 16:12:35 lennart Exp $
#
# (c) 2001, Torped Strategi och Kommunikation AB.
# See License.txt for License information.
#

__doc__="""EasyImage Object"""
__version__='$Revision: 1.18 $'[11:-2]


import Globals
from Globals import DTMLFile
import string
from OFS.Image import Image, Pdata, getImageInfo
try:
    from Shared.Torped import Tutils
    from Products.EasyWorkflow.EasyWorkflowBase import NullDate
    from Products.EasyResources.EasyResource import EasyResource
    from Products.EasyWorkflow.EasyWorkflowBase import NullDate
    from Products.EasyDublinCore.EasyDublinCore import EasyDublinCore
    from Products.EasySecurity.EasyCorePermissions import *
    _EP_is_present = 1
    class ImageClassProxy(EasyResource, EasyDublinCore, Image):
        pass
except: 
    # Running in a  non EasyPublisher environment:
    _EP_is_present = 0
    EasyImageAdd = 'Easy Image Add'
    class ImageClassProxy(Image):
        pass

    
from ComputedAttribute import ComputedAttribute
from webdav.common import rfc1123_date
import os
try:
    import PIL
    import PIL.PngImagePlugin
    import PIL.ImageFile
    PIL.ImageFile.MAXBLOCK = 1000000 # default is 64k
    from StringIO import StringIO
    _PIL_is_present = 1
except:
    _PIL_is_present = 0

# This maps a content type to an internal type.
CONTENTTYPEMAP = { 'image/pjpeg':   'JPEG',     #Progressive JPEG
                   'image/jpeg':    'JPEG', 
                   'image/bmp':     'BMP',  
                   'image/png':     'PNG',
                   'image/gif':     'GIF',
                   'image/tif':     'TIFF',
                 } 

# This contains info on each internal type. (display-name, extension, content_type)
TYPEINFOMAP =  { 'JPEG':     { 'name':'JPEG', 'content_type': 'image/jpeg', }, 
                 'GIF':      { 'name':'GIF',  'content_type': 'image/gif', }, 
                 'BMP':      { 'name':'BMP',  'content_type': 'image/bmp', }, 
                 'PNG':      { 'name':'PNG',  'content_type': 'image/png', }, 
                 'TIFF':     { 'name':'TIFF', 'content_type': 'image/tif', }, 
               }
                 
# This map show possible conversions from each internal type:
CONVERSIONMAP = { 'JPEG': ('PNG', 'TIFF', 'BMP'),
                  'GIF':  ('JPEG', 'PNG', 'TIFF', 'BMP'),
                  'BMP':  ('JPEG', 'PNG', 'TIFF',),
                  'PNG':  ('JPEG', 'TIFF', 'BMP'),
                  'TIFF': ('JPEG', 'PNG', 'BMP',),
                }
                   

class EasyImage(ImageClassProxy):
    """A EasyImage Object """

    meta_type = 'EasyImage'

    description = ""
    keywords = ""
    category = ""

    manage_options_image = ({'label':'Edit',            'action': 'manage_main'       },
                            {'label':'View/Download',   'action': 'view_image_or_file'}, )
    
    manage_options_convertresize = ({'label':'Convert/Resize',  'action': 'manage_convertResizeForm'},)

    manage_options_inherited =  Image.manage_options[2:]

    if _PIL_is_present == 1:
        manage_options = manage_options_image + manage_options_convertresize + manage_options_inherited
        PIL_is_present = 1
    else:
        manage_options = manage_options_image + manage_options_inherited
        PIL_is_present = 0

    def __init__(self, id, file, title=''):
        if _EP_is_present: EasyDublinCore.__init__(self)
        self.originaldata = None
        self.has_preview = 0
        Image.__init__(self, id, title, file)


    def getImageInfoPIL(self, data):
        file = StringIO(str(data))
        im = PIL.Image.open(file)
        size = im.size
        ct = TYPEINFOMAP[im.format]['content_type']
        return ct, size[0], size[1]
    
    
    def update_data(self, data, content_type=None, size=None):
        if size is None: size=len(data)

        self.size=size
        self.data=data

        if _PIL_is_present == 0:
                ct, width, height = getImageInfo(data)
        else:
                ct, width, height = self.getImageInfoPIL(data)        
        
        if ct:
            content_type = ct
        if width >= 0 and height >= 0:
            self.width = width
            self.height = height

        # Now we should have the correct content type, or still None
        if content_type is not None: self.content_type = content_type

        self.ZCacheable_invalidate()

    def manage_upload(self,file='',REQUEST=None):
        """
        Replaces the current contents of the File or Image object with file.

        The file or images contents are replaced with the contents of 'file'.
        """
        Image.manage_upload(self, file, REQUEST)
        
        self.orig_data = None
        self.orig_height = self.orig_width = self.orig_size = 0
        self.prev_data = None
        self.prev_height = self.prev_width = self.prev_size = 0
        
        if self.has_preview:
            create_prev = 1 # Create a preview if there already is one.
        else: # Check for defaults  
            ER = getattr(self, 'EasyResources', None)
            if ER == None:
                create_prev = 1
            else:
                create_prev = getattr(ER, 'default_preview_create', 1)

        if create_prev:
            self.manage_create_prev( ratio=1)
        else:
            self.has_preview = 0
        
        if REQUEST is not None:
            message = "Image Uploaded"
            return self.manage_main(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Edit')
        
        

    def manage_create_prev(self, maxx=0, maxy=0, ratio=0, REQUEST=None):
        """ Creates the preview."""

        if not _PIL_is_present:
            if REQUEST is not None:
                message = "Image library not installed"
                return self.manage_main(
                    self,
                    REQUEST,
                    manage_tabs_message=message,
                    management_view='Edit')
            return 0

        maxx = int(maxx)
        maxy = int(maxy)
        
        if maxx == 0 or maxy == 0:
            # Find the default preview setting, or if it doesn't have one use 100
            defmaxx = defmaxy = 0
            ER = getattr(self, 'EasyResources', None)
            if ER != None:
                defmaxy = getattr(ER, 'default_preview_height', 100)
                defmaxx = getattr(ER, 'default_preview_width', 100)

            if maxx == 0: maxx = defmaxx or 100
            if maxy == 0: maxy = defmaxy or 100

        self.prev_data, self.prev_type, self.prev_width, self.prev_height = \
            self.makeimage( maxx, maxy, ratio)
        self.prev_size = len(self.prev_data)
        if self.prev_size <> 0:
            self.has_preview = 1
            
        if REQUEST is not None:
            message = "Preview created"
            return self.manage_main(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Edit')
        return 1
        
    def manage_del_prev(self, REQUEST=None):
        """Deletes the preview image"""
        self.has_preview = 0
        self.prev_size = 0
        self.prev_type = ''
        self.prev_data = None
        self.prev_height = 0
        self.prev_width = 0
        
        if REQUEST is not None:
            message = "Preview deleted"
            return self.manage_main(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Edit')

    def has_original(self):
        return self.orig_data != None

    def manage_revert(self, REQUEST=None):
        """Reverts to the original image"""
        self.data = self.orig_data
        self.content_type = self.orig_type
        self.size = self.orig_size
        self.width = self.orig_width
        self.height = self.orig_height
        self.orig_size = 0
        self.orig_type = ''
        self.orig_data = None
        self.orig_height = 0
        self.orig_width = 0
        
        if REQUEST is not None:
            message = "Reverted to original"
            return self.manage_main(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Edit')

    
    def set_header(self, REQUEST, RESPONSE, content_type, size):
        # HTTP If-Modified-Since header handling.
        header=REQUEST.get_header('If-Modified-Since', None)
        if header is not None:
            header=string.split(header, ';')[0]
            # Some proxies seem to send invalid date strings for this
            # header. If the date string is not valid, we ignore it
            # rather than raise an error to be generally consistent
            # with common servers such as Apache (which can usually
            # understand the screwy date string as a lucky side effect
            # of the way they parse it).
            try:    mod_since=long(DateTime(header).timeTime())
            except: mod_since=None
            if mod_since is not None:
                if self._p_mtime:
                    last_mod = long(self._p_mtime)
                else:
                    last_mod = long(0)
                if last_mod > 0 and last_mod <= mod_since:
                    # Set header values since apache caching will return Content-Length
                    # of 0 in response if size is not set here
                    RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
                    RESPONSE.setHeader('Content-Type', content_type)
                    RESPONSE.setHeader('Content-Length', size)
                    RESPONSE.setStatus(304)
                    return 304
        return 0

    def index_html(self, REQUEST, RESPONSE, preview=0, original=0):
        """
        The default view of the contents of a File or Image.

        Returns the contents of the file or image.  Also, sets the
        Content-Type HTTP header to the objects content type.
        """
        if preview == 1 and self.has_preview:
            ct = self.prev_type
            size = self.prev_size
            data = self.prev_data
        elif original == 1 and self.orig_data <> None:
            ct = self.orig_type
            size = self.orig_size
            data = self.orig_data
        else:
            ct = self.content_type
            size = self.size
            data = self.data

        self.set_header(REQUEST, RESPONSE, ct, size)

        if self.precondition and hasattr(self,self.precondition):
            # Grab whatever precondition was defined and then 
            # execute it.  The precondition will raise an exception 
            # if something violates its terms.
            c=getattr(self,self.precondition)
            if hasattr(c,'isDocTemp') and c.isDocTemp:
                c(REQUEST['PARENTS'][1],REQUEST)
            else:
                c()

        RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
        RESPONSE.setHeader('Content-Type', ct)
        RESPONSE.setHeader('Content-Length', size)

        # Don't cache the data itself, but provide an opportunity
        # for a cache manager to set response headers.
        self.ZCacheable_set(None)

        if type(data) is type(''): return data

        while data is not None:
            RESPONSE.write(data.data)
            data=data.next

        return ''
    
    
    def tag(self, preview=0, original=0, height=None, width=None, alt=None,
            scale=0, xscale=0, yscale=0, css_class=None, **args):
        """ Generate an HTML IMG tag for this image, with customization."""
        if preview == 1:
            x = self.prev_width
            y = self.prev_height
            u = self.absolute_url() + '/preview'
        elif original == 1:
            x = self.orig_width
            y = self.orig_height
            u = self.absolute_url() + '/original'
        else:
            x = self.width
            y = self.height
            u = self.absolute_url()
        
        if height is None: height=y
        if width is None:  width=x

        # Auto-scaling support
        xdelta = xscale or scale
        ydelta = yscale or scale

        if xdelta and width:
            width =  str(int(round(int(width) * xdelta)))
        if ydelta and height:
            height = str(int(round(int(height) * ydelta)))

        result='<img src="%s"' % (u)

        if alt is None:
            alt=getattr(self, 'title', '')
        result = '%s alt="%s"' % (result, alt)

        if height:
            result = '%s height="%s"' % (result, height)

        if width:
            result = '%s width="%s"' % (result, width)

        if not 'border' in map(string.lower, args.keys()):
            result = '%s border="0"' % result

        if css_class is not None:
            result = '%s class="%s"' % (result, css_class)

        for key in args.keys():
            value = args.get(key)
            result = '%s %s="%s"' % (result, key, value)

        return '%s />' % result

    
    def preview(self, REQUEST, RESPONSE):
        """ return a preview of the image """
        return self.index_html( REQUEST, RESPONSE, preview=1 )
        
    def preview_html(self, **args):
        """Makes an IMG-tag for the preview"""
        args['preview']=1
        return apply(EasyImage.tag, (self,), args)

    def original(self, REQUEST, RESPONSE):
        """ return a preview of the image """
        return self.index_html( REQUEST, RESPONSE, original=1 )

    def original_html(self, **args):
        """Makes an IMG-tag for the original"""
        args['original']=1
        return apply(EasyImage.tag, (self,), args)

    def manage_afterClone(self, item):
        Image.manage_afterClone(self, item)

    def manage_afterAdd(self, item, container):
        Image.manage_afterAdd(self, item, container)
        if hasattr(self, 'index_object'):
            self.index_object()

    def manage_beforeDelete(self, item, container):
        Image.manage_beforeDelete(self, item, container)
        self.unindex_object()

    manage_main = DTMLFile('dtml/manage_editEasyImageForm', globals())

    manage_convertResizeForm = DTMLFile('dtml/manage_convertResizeEasyImageForm', globals())

    def manage_resize_image(self, maxx=0, maxy=0, ratio=0, REQUEST=None):
        """TTW: Resizing images"""
        self.check_original()
        
        if _PIL_is_present == 0:
            if REQUEST is not None:
                message = "Python Imaging Library is not installed. Resizing not supported."
                return self.manage_convertResizeForm(
                    self,
                    REQUEST,
                    manage_tabs_message=message,
                    management_view='Convert/Resize')
            


        self.data, self.content_type, self.width, self.height = self.makeimage(maxx, maxy, ratio, CONTENTTYPEMAP[self.content_type])

        if REQUEST is not None:
            message = "Image Resized."
            return self.manage_convertResizeForm(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Convert/Resize')
        

    
    def check_original(self):
        if self.orig_data == None:
            self.orig_data = self.data
            self.orig_type = self.content_type
            self.orig_size = self.size
            self.orig_width = self.width
            self.orig_height = self.height

    def makeimage(self, maxx=0, maxy=0, ratio=1, ftype='JPEG'):
        """Creates an image from the orginal with specified size."""

        if self.orig_data == None:
            tmpdata = self.data
        else:
            tmpdata = self.orig_data
            
        file = StringIO(str(tmpdata))
                
        if maxx <> 0 and maxy <> 0:
            im = PIL.Image.open(file)

            if ratio==1:            # keep aspect-ratio
                aspectx = float(im.size[0]) / float(maxx)
                aspecty = float(im.size[1]) / float(maxy)
                if aspectx > aspecty:
                    maxy = float(im.size[1]) / aspectx
                if aspectx < aspecty:
                    maxx = float(im.size[0]) / aspecty

            maxx = int(maxx)
            maxy = int(maxy)
            if im.mode!='RGB' and im.mode!='CMYK': im = im.convert("RGB")
            im.load()
            im = im.resize((maxx, maxy), PIL.Image.BICUBIC)
            
            newfile = StringIO()
            im.save(newfile, ftype, optimize=1)
            return newfile.getvalue(), TYPEINFOMAP[ftype]['content_type'], im.size[0], im.size[1]

        return None, '', 0, 0

    def get_conversion_types(self, content_type=None):
        
        if content_type == None: content_type = self.content_type
        try:
            filetype = CONTENTTYPEMAP[content_type]
        except:
            filetype = content_type

        conversiontypes = CONVERSIONMAP[filetype]
        res = []
        for ct in conversiontypes:
            res = res + [(ct, TYPEINFOMAP[ct]['name'])]
        
        return res

    def manage_convert_image(self, new_format='', REQUEST=None):
        """TTW: Converting images"""
        self.check_original() # Make sure the original is kept.

        if new_format<>'':
            if _PIL_is_present == 0:
                if REQUEST is not None:
                    message = "Python Imaging Library is not installed. Conversions not supported."
                    return self.manage_convertResizeForm(
                        self,
                        REQUEST,
                        manage_tabs_message=message,
                        management_view='Convert/Resize')
                
            else:
                self.data, self.content_type, self.width, self.height = self.makeimage(self.width, self.height, ratio=0, ftype=new_format)

        if REQUEST is not None:
            message = "Image converted."
            return self.manage_convertResizeForm(
                self,
                REQUEST,
                manage_tabs_message=message,
                management_view='Convert/Resize')
        


Globals.InitializeClass(EasyImage)

manage_addEasyImageForm=DTMLFile('dtml/manage_addEasyImageForm', globals())

#    """Add EasyImage"""
def manage_addEasyImage(self, id, file, title='', content_type='', REQUEST=None):
    """
    Add a new EasyImage object.

    Creates a new Image object 'id' with the contents of 'file'.
    """
    if not id and hasattr(file,'filename'):
        # generate id from filename and make sure, it has no 'bad' chars
        # TODO: Strip extension.
        id = os.path.splitext(file.filename)[0]
        id = id[max(string.rfind(id,'/'),
                string.rfind(id,'\\'),
                string.rfind(id,':')
               )+1:]
        title = title or string.capitalize(id)
 
    if _EP_is_present:
        id=Tutils.xlate2url(id)
    title=str(title)
    content_type=str(content_type)

    self = self.this()
    self._setObject(id, EasyImage(id, file, title))

    imgObj = self._getOb(id)
    imgObj.manage_upload(file)
    if content_type:
        self._getOb(id).content_type=content_type

    if REQUEST is not None:
        try:    url=self.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    

def initialize(context):
    """Initialize EasyImage Product"""

    context.registerClass (
            EasyImage,
            permission=EasyImageAdd,
            icon='icons/EasyImage.gif',
            constructors=(manage_addEasyImageForm, manage_addEasyImage),
            meta_type = EasyImage.meta_type
            )


# $Log: EasyImage.py,v $
# Revision 1.18  2002/05/08 16:12:35  lennart
# Now supports Zope without EasyPublisher
#
# Revision 1.17  2002/05/08 15:39:58  lennart
# More bugfixes
#
# Revision 1.16  2002/05/08 09:01:03  lennart
# Moved the init of has_preview to before the upload.
#
# Revision 1.15  2002/05/02 12:55:24  lennart
# Most services + Easy Document now refrshes the navigation tree.
#
# Revision 1.14  2002/05/01 12:17:23  lennart
# Removed some debug code
#
# Revision 1.13  2002/05/01 11:49:02  lennart
# Fixed the handling of EasyResource deafult settings.
#
# Revision 1.12  2002/04/30 19:46:07  lennart
# Saves the original image on resize.
#
# Revision 1.11  2002/04/30 19:38:31  lennart
# Added showing of original data and a revert button + some small fixes.
#
# Revision 1.10  2002/04/28 15:52:49  johanc
# Ported to Zope 2.5.1
#
# Revision 1.9  2002/04/16 08:20:31  johanc
# Fixed upload form that didn't work.
#
# Revision 1.8  2002/03/22 21:26:59  lennart
# Now uses Shared.Torped.Tutils.xlate2url fo url safe id's.
#
# Revision 1.7  2002/03/08 10:47:15  zhost
# Added tag parameters (like align, width, height) to preview_html and original_html.
#
# Revision 1.6  2002/02/18 10:38:26  lennart
# Fixed bug in the handling of PIL enabled environments.
#
# Revision 1.5  2002/01/06 11:05:30  lennart
# *** empty log message ***
#
# Revision 1.4  2002/01/03 18:19:06  lennart
# Cleanup of path attributes and function names
#
# Revision 1.3  2002/01/03 14:46:30  lennart
# More binary fixes
#
# Revision 1.2  2002/01/03 12:31:46  lennart
# Minor bugs
#
# Revision 1.6  2002/01/03 10:39:15  lennart
# Trying again
#
